/****************************************************************************
*
*     Copyright (c) 2005 Dave Hylands
*           All Rights Reserved
*
*   Permission is granted to any individual or institution to use, copy, or
*   redistribute this software so long as it is not sold for profit, and that
*   this copyright notice is retained.
*
****************************************************************************/
/**
*
*   @file   BootHost.cpp
*
*   @brief  Implements the i2c BootLoader Host
*
****************************************************************************/
/**
*   @defgroup   BootHost   i2c BootLoader Host.
*
*   @brief      BootLoad Host program for downloading to target via i2c.
*
****************************************************************************/

// ---- Include Files -------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/unistd.h>
#include <pthread.h>
#include <getopt.h>
#include <termios.h>

#include "BootLoader.h"
#include "SerialPort.h"
#include "ReadFile.h"
#include "State.h"
#include "StateAVReflash.h"
#include "StateMegaLoad.h"
#include "Log.h"

#include "StateSTK500.h"
#include "stk500-command.h"

#include "I2C_SerialDongle.h"

// ---- Public Variables ----------------------------------------------------
// ---- Private Constants and Types -----------------------------------------

// ---- Private Variables ---------------------------------------------------

struct option gLongOption[] =
{
    // option       A  Flag   V  (has_arg, flag, val)
    // -----------  -  ----  ---
    { "baud",       1, NULL, 'b' },
    { "debug",      0, NULL, 'd' },
    { "dongle",     0, NULL, 'g' }, // Use Serial <==> I2C dongle
    { "help",       0, NULL, 'h' },
    { "megaload",   0, NULL, 'm' }, // Use MegaLoad protocol
    { "port",       1, NULL, 'p' },
    { "stk500",     0, NULL, 's' }, // Use STK500 protocol
    { "verbose",    0, NULL, 'v' },
    { 0 },

};

bool gMegaLoad = false;
bool gStk500 = true;
bool gDongle = false;
bool gBootLoadAllowed = true;

// We can program the serial <=> i2c dongle, which is connected to the
// serial port, or we can program a device on the i2c bus. To avoid confusion
// we turn off some reporting if we don't intend to program the dongle.

bool gProgramDongle = true;

SerialPort          gSerialPort;
I2C_SerialDongle    gSerialDongle;
unsigned char       gSerialDongleAddr = 0;

DeviceInfo      gDeviceInfo;
const char     *gDownloadFileName = NULL;
FileInfo       *gDownloadInfo = NULL;
unsigned        gRetryCount = 0;
unsigned char   gSlaveAddr = 0x34;
int             gCounter = 111;

// Used by 
FileSegment    *gDownloadSegment = NULL;
unsigned        gDownloadOffset;

// ---- Private Function Prototypes -----------------------------------------

void DumpMem( const char *prefix, unsigned address, const void *data, unsigned numBytes );

void   *ReaderThread( void *param );
void    Usage();

void LogUnrecognizedChar( char ch, const char *label );

void PacketTextChar( unsigned char ch, void *userData );
void PacketSendChar( unsigned char ch, void *userData );
void PacketReceived( PKT_Packet *packet, void *userData );

void SendByte( unsigned char ch );

// ---- Functions -----------------------------------------------------------

/**
 * @addtogroup BootHost
 * @{
 */

int main( int argc, char **argv )
{
    pthread_t   readerThreadId;
    int         rc;
    int         opt;
    const char *baudStr = NULL;
    const char *portStr = "com1";


    PKT_TextChar        = PacketTextChar;
    PKT_SendChar        = PacketSendChar;
    PKT_PacketReceived  = PacketReceived;

    LogInit( stdout );

    while (( opt = getopt_long( argc, argv, "b:dhmp:sv", gLongOption, NULL )) > 0 )
    {
        switch ( opt )
        {
            case 'b':
            {
                baudStr = optarg;
                break;
            }

            case 'd':
            {
                gDebug = true;
                break;
            }

            case 'g':
            {
                gDongle = true;
                break;
            }

            case 'm':
            {
                gMegaLoad = true;
                break;
            }

            case 'p':
            {
                portStr = optarg;
                break;
            }

            case 's':
            {
                gStk500 = true;
                break;
            }

            case 'v':
            {
                gVerbose = true;
                break;
            }
            case '?':
            case 'h':
            {
                Usage();
                return 1;
            }
        }
    }

    if ( optind < argc )
    {
        if (( optind + 1 ) != argc )
        {
            fprintf( stderr, "Only one download file supported\n" );
            return 1;
        }

        gDownloadFileName = argv[ optind ];
    }

    // Open the file to download

    if ( gDownloadFileName != NULL )
    {
        // If we are asked to download a file, then read the entire file
        // into memory.

        if (( gDownloadInfo = ReadFile( gDownloadFileName )) == NULL )
        {
            return 1;
        }
    }

    if ( !gSerialPort.Open( portStr, baudStr ))
    {
        return 1;
    }

    // Put stdin in raw mode

    setbuf( stdin, NULL );

    struct termios tio_org;
    struct termios tio_new;
    
    if ( ioctl( fileno( stdin ), TCGETA, &tio_org ) == -1 )
    {
        LogError( "Unable to retrieve terminal settings\n" );
        return 1;
    }
    
    tio_new = tio_org;
    tio_new.c_lflag &= ~( ICANON | ECHO );
    tio_new.c_cc[VMIN] = 1;
    tio_new.c_cc[VTIME] = 0;
    
    if ( ioctl( fileno( stdin ), TCSETA, &tio_new ) == -1 )
    {
        LogError( "Unable to update terminal settings\n" );
        return 1;
    }

    Log( "BootHost MegaLoad:%d Debug:%d Verbose:%d\n", gMegaLoad, gDebug, gVerbose );


    // Kick off the serial port reader thread.

    rc = pthread_create( &readerThreadId, NULL, ReaderThread, &gSerialPort );
    if ( rc != 0 )
    {
        fprintf( stderr, "Error creating ReaderThread: %d\n", rc );
        return 1;
    }

    while ( 1 )
    {
        int ch = fgetc( stdin );

        if ( gDongle )
        {
            if ( gSerialDongleAddr == 0 )
            {
                if ( !gSerialDongle.GetI2CAddress( &gSerialDongleAddr ))
                {
                    LogError( "Unable to retrieve serial dongle i2c address\n" );
                    return 1;
                }
                LogVerbose( "Serial Dongle i2c Address: 0x%02x\n", gSerialDongleAddr );
            }

            switch ( ch )
            {
                case 'c':
                {
                    printf( "Call\n" );

                    I2C_Adapter::Buffer  writeBuf;
                    I2C_Adapter::Buffer  readBuf;

                    char    writeMem[ 40 ];
                    char    readMem[ 40 ];

                    sprintf( writeMem, "-c-%d-c-", gCounter++ );

                    writeBuf.data = (unsigned char *)writeMem;
                    writeBuf.dataLen = strlen( writeMem ) + 1;

                    readBuf.data = (unsigned char *)readMem;
                    readBuf.dataLen = 0;

                    gSerialDongle.Call( gSlaveAddr, BL_CMD_TEST_CALL, &writeBuf, &readBuf, sizeof( readMem ));

                    printf( "Call: Wrote: '%s', read '%s'\n", writeMem, readMem );
                    break;
                }

                case 'r':
                {
                    printf( "Read\n" );

                    I2C_Adapter::Buffer readBuf;
                    char                readMem[ 40 ];

                    readBuf.data = (unsigned char *)readMem;
                    readBuf.dataLen = 0;

                    gSerialDongle.Read( gSlaveAddr, BL_CMD_TEST_READ, &readBuf, sizeof( readMem ));

                    printf( "Read: '%s'\n", readMem );
                    break;
                }

                case 'w':
                {
                    printf( "Write\n" );

                    I2C_Adapter::Buffer writeBuf;
                    char                writeMem[ 40 ];

                    sprintf( writeMem, "-w-%d-w-", gCounter++ );

                    writeBuf.data = (unsigned char *)writeMem;
                    writeBuf.dataLen = strlen( writeMem ) + 1;

                    printf( "Call: Wrote: '%s'\n", writeMem );

                    gSerialDongle.Write( gSlaveAddr, BL_CMD_TEST_WRITE, &writeBuf );

                    break;
                }

                default:
                {
                    gSerialPort.Write( &ch, 1 );
                    break;
                }
            }
        }
        else
        {
            gSerialPort.Write( &ch, 1 );
        }
    }

    return 0;

} // main

/***************************************************************************/
/**
*   Checks the file information to see if the file will fit on the device.
*/

bool CheckFile( FileInfo *fileInfo, DeviceInfo *devInfo )
{
    unsigned    devEndAddr;

    devEndAddr = devInfo->flashSize - devInfo->bootSize;

    FileSegment *segment = fileInfo->head;
    while ( segment != NULL )
    {
        if (( segment->address + segment->lenThisSegment ) > devEndAddr )
        {
            LogError( "File '%s' is too big to fit\n", fileInfo->fileName );
            return false;
        }
        segment = segment->next;
    }

    return true;

} // CheckFile

/**************************************************************************/
/**
*   Dumps a page of output for debugging purposes.
*/

void DumpMem( const char *prefix, unsigned address, const void *inData, unsigned numBytes )
{
    const uint8_t   *data = (const uint8_t *)inData;

#define LINE_WIDTH  16

    for ( unsigned byteOffset = 0; byteOffset < numBytes; byteOffset += LINE_WIDTH ) 
    {
        unsigned    i;

        Log( "%s: %04x: ", prefix, address + byteOffset );

        for ( i = 0; i < LINE_WIDTH; i++ ) 
        {
            if (( byteOffset + i ) < numBytes )
            {
                Log( "%02.2X ", data[ byteOffset + i ] );
            }
            else
            {
                Log( "   " );
            }
        }
        for ( i = 0; i < LINE_WIDTH; i++ ) 
        {
            if (( byteOffset + i ) < numBytes )
            {
                unsigned char ch = data[ byteOffset + i ];
                if (( ch < ' ' ) || ( ch > '~' ))
                {
                    Log( "." );
                }
                else
                {
                    Log( "%c", ch );
                }
            }
            else
            {
                break;
            }
        }
        Log( "\n" );
    }

} // DumpMem

/***************************************************************************/
/**
*   Logs an unrecognized character while processing in the state machine
*/

void LogUnrecognizedChar( char ch, const char *label )
{
    LogError( "Unrecognized character received: 0x%02x while waiting for %s\n", ch, label );

} // LogUnrecognizedChar

/***************************************************************************/
/**
*   Callback which is called when a text char is received on the serial port.
*/

void PacketTextChar( unsigned char ch, void *userData )
{
    (void)userData;

    Log( "\033[1;32m%c\033[0m", ch );

} // PacketTextChar

/***************************************************************************/
/**
*   Callback which is called when a text char is received on the serial port.
*/

void PacketSendChar( unsigned char ch, void *userData )
{
    if ( gDebug )
    {
        Log( "Sending: 0x%02x\n", ch );
    }
    SendByte( ch );

} // PacketSendChar

/***************************************************************************/
/**
*   Callback which is called when a packet is received on the serial port.
*/

void PacketReceived( PKT_Packet *packet, void *userData )
{
    gSerialDongle.PacketReceived( packet );

} // PacketReceived

/***************************************************************************/
/**
*   Thread which processes the incoming serial data.
*/

void *ReaderThread( void *param )
{
    SerialPort *serialPort = (SerialPort *)param;

    char    buf[ 256 ];

    StatePtrProxy   state = NULL;

    Log( "ReaderThread started\n" );

    if ( gStk500 )
    {
        state = StartSTK500();
    }
    else
    {
        state = StateDownload;
    }

    while ( 1 )
    {
        size_t  bytesRead = serialPort->Read( buf, sizeof( buf ));

        for ( unsigned i = 0; i < bytesRead; i++ ) 
        {
#if 1
            if ( gDebug )
            {
                unsigned char ch = buf[ i ];

                if (( ch < ' ' ) || ( ch > '~' ))
                {
                    Log( "BH-Rcvd: 0x%02x '.'\n", ch );
                }
                else
                {
                    Log( "BH-Rcvd: 0x%02x '%c'\n", ch, ch );
                }

            }
#endif

            state = state( buf[ i ]);
            if ( state == NULL )
            {
                state = StateLog;
            }
        }
    }

    return 0;

} // ReaderThread

/***************************************************************************/
/**
*   Sends a character out the serial port.
*/

void SendByte( unsigned char ch )
{
    if ( gDebug )
    {
        Log( "Sending: 0x%02x\n", ch );
    }
    gSerialPort.Write( &ch, 1 );

} // SendByte

void SendData( const void *buf, size_t bytesToWrite )
{
    if ( gDebug )
    {
        DumpMem( "Send", 0, buf, bytesToWrite );
    }
    gSerialPort.Write( buf, bytesToWrite );

} // SendData

/***************************************************************************/
/**
*   Normal starting state. We log each character received, unless it looks
*   like the start of a packet, or a MegaLoad boot request.
*/

StatePtrProxy StateDownload( char ch )
{
    if ( gBootLoadAllowed )
    {
        if ( ch == '>' )
        {
            SendByte( '<' );
            gBootLoadAllowed = false;
            memset( &gDeviceInfo, 0, sizeof( gDeviceInfo ));

            if ( gMegaLoad )
            {
                return StateMegaDeviceId;
            }

            return StateAVReflashPageSize;
        }

        // Log the character, but stay in the download state

        StateLog( ch );

        return StateDownload;
    }

    // Switch over to logging

    return StateLog( ch );

} // StateDownload

/***************************************************************************/
/**
*   Just performs logging.
*/

StatePtrProxy StateLog( char ch )
{
    PKT_ProcessChar( ch, NULL );

    return StateLog;

} // StateLog

/***************************************************************************/
/**
*   Print out the program usage.
*/

void Usage()
{
    fprintf( stderr, "Usage: BootHost <option(s)> <file>\n" );
    fprintf( stderr, "  Download a program via serial/i2c\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "  -b, --baud=baud   Set the baudrate used\n" );
    fprintf( stderr, "  -h, --help        Display this message\n" );
    fprintf( stderr, "  -m, --megaload    Use megaload 2.3 protocol\n" );
    fprintf( stderr, "  -s, --stk500      Use stk500 bootloader protocol (default)\n" );
    fprintf( stderr, "  -v, --verbose     Turn on verbose messages\n" );

} // Usage

/** @} */

