/****************************************************************************
*
*     Copyright (c) 2005 Dave Hylands
*           All Rights Reserved
*
*   Permission is granted to any individual or institution to use, copy, or
*   redistribute this software so long as it is not sold for profit, and that
*   this copyright notice is retained.
*
****************************************************************************/
/**
*
*   @file   I2C_Adapter.h
*
*   @brief  This file describes the interface to the I2C_Adapter class.
*
****************************************************************************/
/**
*   @defgroup   i2c   i2c abstract base class.
*
*   @brief      Interface to the i2c device.
*
****************************************************************************/

#if !defined( I2C_ADAPTER_H )
#define I2C_ADAPTER_H                   ///< Include Guard

// ---- Include Files -------------------------------------------------------

#include <stdlib.h>

/**
 * @addtogroup i2c
 * @{
 */

//---------------------------------------------------------------------------
/**
*   The I2C_Adapter class forms an abstract base class, used to communicate with
*   i2c devices. A derived class does the real work. This way we can have
*   a common interface for say talking directly with i2c (on the gumstix) or
*   through an i2c serial dongle (oin the PC).
*/

class I2C_Adapter
{
public:

    struct Buffer
    {
        unsigned char   dataLen;
        unsigned char  *data;
    };

    class Callback
    {
    public:
        virtual ~Callback();
        virtual void PacketReceived( Buffer *buf ) = 0;
    };

    //------------------------------------------------------------------------
    // The ReadWrite flag is or'd together with the slave address.

    enum SlaveReadWrite
    {
        SlaveWrite   = 0x00,
        SlaveRead    = 0x01,
    };

    //------------------------------------------------------------------------
    //  Default constructor

    I2C_Adapter();

    //------------------------------------------------------------------------
    //  Destructor

    virtual ~I2C_Adapter();

    //------------------------------------------------------------------------
    // Sets the callback which is called when a packet is received

    void SetCallback( Callback *callback );

    //------------------------------------------------------------------------
    // Retrieves the i2c address associated with the serial dongle

    virtual bool GetI2CAddress( unsigned char *masterAddr ) = 0;

    //------------------------------------------------------------------------
    //  Reads data from an i2c slave.
    //
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   Repeated Start
    //  M->S:   slave-addr w/ R
    //  S->M:   len
    //  S->M:   data (first byte should be address of slave)
    //  S->M:   CRC
    //  M->S:   Stop

    virtual bool Read( unsigned char slaveAddr, unsigned char cmd, Buffer *buf, size_t maxLen ) = 0;

    //------------------------------------------------------------------------
    //  Writes data to an i2c slave.
    //
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   len
    //  M->S:   data (first byte should be address of master)
    //  M->S:   CRC
    //  M->S:   Stop

    virtual bool Write( unsigned char slaveAddr, unsigned char cmd, const Buffer *buf )= 0;

    //------------------------------------------------------------------------
    //  Does a Write/Read (aka Process Call) with an i2c slave
    //
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   len
    //  M->S:   data (first byte should be address of master)
    //  M->S:   Repeated Start
    //  M->S:   slave-addr w/ R
    //  S->M:   len
    //  S->M:   data (first byte should be address of slave)
    //  S->M:   CRC
    //  M->S:   Stop

    virtual bool Call( unsigned char slaveAddr, unsigned char cmd, const Buffer *writeBuf, Buffer *readBuf, size_t maxLen ) = 0;

private:

    //------------------------------------------------------------------------
    // The copy constructor and assignment operator are not need for this
    // class so we declare them private and don't provide an implementation.

    I2C_Adapter( const I2C_Adapter & copy );
    I2C_Adapter &operator =( const I2C_Adapter &rhs );

    //------------------------------------------------------------------------

protected:

    Callback   *m_callback;         ///< Fucntion called when the read data arrives.
};

// ---- Inline Functions ----------------------------------------------------

/** @} */

#endif // I2C_ADAPTER_H

