/****************************************************************************
*
*     Copyright (c) 2005 Dave Hylands
*           All Rights Reserved
*
*   Permission is granted to any individual or institution to use, copy, or
*   redistribute this software so long as it is not sold for profit, and that
*   this copyright notice is retained.
*
****************************************************************************/
/**
*
*   @file   I2C_SerialDongle.cpp
*
*   @brief  This file implemented the I2C_SerialDongle device.
*
****************************************************************************/

// ---- Include Files -------------------------------------------------------

#include "I2C_SerialDongle.h"
#include "PKT_Packet.h"

#include "Crc8.h"
#include "Log.h"

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <semaphore.h>
#include <time.h>

// ---- Public Variables ----------------------------------------------------
// ---- Private Constants and Types -----------------------------------------
// ---- Private Variables ---------------------------------------------------
// ---- Private Function Prototypes -----------------------------------------
// ---- Functions -----------------------------------------------------------

/**
 * @addtogroup i2c
 * @{
 */

//***************************************************************************
/**
*   Constructor
*/

I2C_SerialDongle::I2C_SerialDongle()
{
    // We assume that the caller has installed callbacks with the 
    // PacketProtocol module.

    memset( &m_response, 0, sizeof( m_response ));

    if ( sem_init( &m_response.m_semaphore, 0, 0 ) != 0 )
    {
        LogError( "I2C_SerialDongle: call to sem_init failed: %s (%d)\n", strerror( errno ), errno );
        m_response.m_semaphore = 0;
    }

    m_dongleAddr = 0;
}

//***************************************************************************
/**
*   Destructor
*
*   virtual
*/

I2C_SerialDongle::~I2C_SerialDongle()
{
    if ( m_response.m_semaphore > 0 )
    {
        sem_destroy( &m_response.m_semaphore );
    }
}

//***************************************************************************
/**
*   Retrieves the I2C address associated with the serial dongle.
*
*   virtual
*/

bool I2C_SerialDongle::GetI2CAddress( unsigned char *masterAddr )
{
    PKT_Packet  pkt;

    pkt.data[ 0 ] = GetI2CAddrPacket;
    pkt.len = 1;

    PKT_SendPacket( &pkt, this );

    Buffer response;

    response.dataLen = 0;
    response.data = masterAddr;

    // Since this is a synchronous call, we need to wait for the data to 
    // arrive.

    if ( WaitForResponse( GetI2CAddrReplyPacket, 0, &response, 1 ))
    {
        m_dongleAddr = *masterAddr;
        return true;
    }
    return false;
    
} // GetI2CAddress

//***************************************************************************
/**
*   Does a Write/Read (aka Process Call) with an i2c slave
*
*   virtual
*/

bool I2C_SerialDongle::Call
(
    unsigned char slaveAddr,    ///< I2C Address of the slave device
    unsigned char cmd,          ///< Command (type of call being performed)
    const Buffer *writeBuf,     ///< Buffer of data to send down to the device
    Buffer        *readBuf,     ///< Buffer to place results in
    size_t        maxLen        ///< Maximum size of buffer (in bytes)
)
{

    //------------------------------------------------------------------------
    //  Does a Write/Read (aka Process Call) with an i2c slave
    //
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   len
    //  M->S:   data (first byte should be address of master)
    //  M->S:   Repeated Start
    //  M->S:   slave-addr w/ R
    //  S->M:   len
    //  S->M:   data (first byte should be address of slave)
    //  S->M:   CRC
    //  M->S:   Stop

    PKT_Packet  pkt;

    unsigned char len = writeBuf->dataLen;
    if ( len > ( sizeof( pkt.data ) - 5 ))
    {
        LogError( "I2C_SerialDongle::Call len (%d) is too big for packet (%d)\n", len, sizeof( pkt.data ));

        len = sizeof( pkt.data ) - 5;
    }

    unsigned char  *pd = &pkt.data[ 0 ];
    unsigned char   crc = 0;

    *pd++ = CallPacket;     // Not sent on i2c bus, therefore not included in crc

    *pd = ( slaveAddr << 1 ) | SlaveWrite;  crc = Crc8( crc, *pd );    pd++;
    *pd = cmd;                              crc = Crc8( crc, *pd );    pd++;
    *pd = writeBuf->dataLen + 1;            crc = Crc8( crc, *pd );    pd++;
    *pd = m_dongleAddr;                     crc = Crc8( crc, *pd );    pd++;

    for ( int i = 0; i < len; i++ ) 
    {
        *pd = writeBuf->data[ i ];
        crc = Crc8( crc, *pd );
        pd++;
    }
    *pd = ( slaveAddr << 1 ) | SlaveRead;   crc = Crc8( crc, *pd );    pd++;

    pkt.len = pd - &pkt.data[ 0 ];

    PKT_SendPacket( &pkt, this );

    // Since this is a synchronous call, we need to wait for the data to 
    // arrive.

    return WaitForResponse( CallReplyPacket, crc, readBuf, maxLen );

} // Call

//***************************************************************************
/**
*   Reads data from an i2c slave.
*
*   virtual
*/

bool I2C_SerialDongle::Read
(
    unsigned char slaveAddr,    ///< I2C Address of the slave device
    unsigned char cmd,          ///< Command (type of read being performed)
    Buffer       *buf,          ///< Buffer to place results in
    size_t        maxLen        ///< Maximum size of buffer (in bytes)
)
{
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   Repeated Start
    //  M->S:   slave-addr w/ R
    //  S->M:   len
    //  S->M:   data (first byte should be address of slave)
    //  S->M:   CRC
    //  M->S:   Stop

    PKT_Packet  pkt;

    unsigned char  *pd = &pkt.data[ 0 ];
    unsigned char   crc = 0;

    *pd++ = ReadPacket;     // Not send on i2c bus, therefore not included in crc

    *pd = ( slaveAddr << 1 ) | SlaveWrite;  crc = Crc8( crc, *pd );    pd++;
    *pd = cmd;                              crc = Crc8( crc, *pd );    pd++;
    *pd = ( slaveAddr << 1 ) | SlaveRead;   crc = Crc8( crc, *pd );    pd++;

    pkt.len = pd - &pkt.data[ 0 ];

    PKT_SendPacket( &pkt, this );

    // Since this is a synchronous call, we need to wait for the data to 
    // arrive.

    return WaitForResponse( ReadReplyPacket, crc, buf, maxLen );

} // Read

//***************************************************************************
/**
*   Writes data to an i2c slave.
*
*   virtual
*/

bool I2C_SerialDongle::Write
(
    unsigned char slaveAddr,    ///< I2C Address of the slave device
    unsigned char cmd,          ///< Command (type of write being performed)
    const Buffer *buf           ///< Buffer of data to send down to the device
)
{
    //  On the i2c bus, the following will take place (M=Master, S=Slave):
    //
    //  M->S:   Start
    //  M->S:   slave-addr w/ W
    //  M->S:   command
    //  M->S:   len
    //  M->S:   data (first byte should be address of master)
    //  M->S:   CRC
    //  M->S:   Stop

    PKT_Packet  pkt;

    unsigned char len = buf->dataLen;
    if ( len > ( sizeof( pkt.data ) - 4 ))
    {
        LogError( "I2C_SerialDongle::Write len (%d) is too big for packet (%d)\n", len, sizeof( pkt.data ));

        len = sizeof( pkt.data ) - 4;
    }

    unsigned char  *pd = &pkt.data[ 0 ];
    unsigned char   crc = 0;

    *pd++ = WritePacket;     // Not send on i2c bus, therefore not included in crc

    *pd = ( slaveAddr << 1 ) | SlaveWrite;  crc = Crc8( crc, *pd );    pd++;
    *pd = cmd;                              crc = Crc8( crc, *pd );    pd++;
    *pd = buf->dataLen + 1;                 crc = Crc8( crc, *pd );    pd++;
    *pd = m_dongleAddr;                     crc = Crc8( crc, *pd );    pd++;

    for ( int i = 0; i < len; i++ ) 
    {
        *pd = buf->data[ i ];
        crc = Crc8( crc, *pd );
        pd++;
    }
    *pd++ = crc;

    pkt.len = pd - &pkt.data[ 0 ];

    PKT_SendPacket( &pkt, this );

    return true;

} // Write

//***************************************************************************
/**
*   Function we expect to be called when a packet is received over the
*   serial port. Note that it is expected that this function will be
*   called from a different thread than the one which calls Read or Call.
*/

void I2C_SerialDongle::PacketReceived( PKT_Packet *packet )
{
    if ( packet->len == 0 )
    {
        LogError( "I2C_SerialDongle::PacketReceived: empty packet\n" );
        return;
    }

    unsigned char *pd = &packet->data[ 0 ];

    PacketType  packetType = (PacketType)*pd++;

    unsigned char   crc = 0;

    if ( packetType == m_response.m_responseType )
    {
        // This is the type of packet that the caller is waiting for.

        crc = m_response.m_partialCrc;
    }

    // Check the CRC to see if it's valid

    unsigned char len = *pd;    crc = Crc8( crc, *pd ); pd++;

    // Sanity check the length. The packet length should equal the i2c len
    // plus 3 (1 for packet type, 1 for len, 1 for CRC)

    if (( len + 3 ) != packet->len )
    {
        LogError( "I2C_SerialDongle::PacketReceived Inconsistent packet length. Serial packetLen = %d, i2c packetLen = %d (expecting %d)\n",
                  packet->len, len, packet->len - 3 );
        return;
    }

    for ( int i = 0; i < len; i++ ) 
    {
        crc = Crc8( crc, pd[ i ]);
    }
    if ( crc != pd[ len ] )
    {
        LogError( "I2C_SerialDongle::PacketReceived CRC failure. Found 0x%02x, expecting 0x%02x\n", pd[ len ], crc );
        return;
    }

    if ( packetType == m_response.m_responseType )
    {
        // We've received a good packet, and we're waiting for this type
        // of packet. Wakeup the reader

        if ( len < 1 )
        {
            LogError( "I2C_SerialDongle::PacketReceived packet len too small (%d). Needs to be at least 1 to store address of sender\n", len );
            len = 1;
        }
        len--;  // We hide the first character of the data, which is the i2c address of the sender

        if ( len > m_response.m_maxLen )
        {
            LogError( "I2C_SerialDongle::PacketReceived packet buffer too small. maxLen = %d, received %d\n", m_response.m_maxLen, len );

            len = m_response.m_maxLen;
        }
        m_response.m_addr = pd[ 0 ];

        m_response.m_buf->dataLen = len;
        if ( len > 0 )
        {
            memcpy( &m_response.m_buf->data[ 0 ], &pd[ 1 ], len );
        }

        // Setting m_responseType to NoPacket signals that we're no longer
        // waiting for a packet.

        m_response.m_responseType = NoPacket;

        sem_post( &m_response.m_semaphore );
        return;
    }

    // Call the registered callback to report the received packet

    m_callback->PacketReceived( m_response.m_buf );

} // PacketReceived

//***************************************************************************
/**
*   Called to block and wait for a response. Returns false if a timeout
*   occurs.
*/
bool I2C_SerialDongle::WaitForResponse( PacketType responseType, unsigned char partialCrc, Buffer *buf, unsigned char maxLen )
{
    // Currently we only handle one outstanding "wait" at a time.

    assert( m_response.m_responseType == NoPacket );

    m_response.m_responseType = responseType;
    m_response.m_partialCrc = partialCrc;
    m_response.m_maxLen = maxLen;
    m_response.m_addr = 0;
    m_response.m_buf = buf;

    struct timespec waitTime;

    waitTime.tv_sec = time( NULL ) + 3;    // We shouldn't have to wait this long for a response
    waitTime.tv_nsec = 0;

    printf( "About to wait\n" );
    if ( sem_timedwait( &m_response.m_semaphore, &waitTime ) != 0 )
    {
        printf( "Finished waiting\n" );
        if ( errno == ETIMEDOUT )
        {
            LogError( "I2C_Dongle::WaitForResponse timeout waiting for response\n" );
        }
        else
        {
            LogError( "I2C_Dongle::WaitForResponse: error waiting for semaphore: %s (%d)\n", strerror( errno ), errno );
        }
        return false;
    }

    return true;

} // WaitForResponse

/** @} */

