/****************************************************************************
*
*     Copyright (c) 2004 Dave Hylands
*           All Rights Reserved
*
*   Permission is granted to any individual or institution to use, copy, or
*   redistribute this software so long as it is not sold for profit, and that
*   this copyright notice is retained.
*
****************************************************************************/
/**
*
*   @file   BootLoader.h
*
*   @brief  Contains the definitions used to communicate with the 
*           Bootloder.
*
****************************************************************************/
/**
*   @defgroup   BootLoader   I2C BootLoader
*
*   @brief      BootLoader for downloading programs to an AVR via
*               the I2C bus.
*
****************************************************************************/

#if !defined( BOOTLOADER_H )
#define BOOTLOADER_H                   ///< Include Guard

/* ---- Include Files ---------------------------------------------------- */

#include <inttypes.h>

/**
 *  @addtogroup BootLoader
 *  @{
 */

/* ---- Constants and Types ---------------------------------------------- */

#define I2C_DEBUG   1

#if I2C_DEBUG
#   define I2C_LOG( fmt, args... )      Log( fmt, ## args )
#else
#   define I2C_LOG( fmt, args... )
#endif

/**
 *  Defines the version of this API. This includes the layout of the 
 *  various structures, along with the semantics associated with the 
 *  protocol. Any changes require the version number to be incremented.
 */

#define BL_API_VERSION      1

/**
 *  The min version, determines the minimum version that this API is
 *  compatable with. This allows old host side programs to determine
 *  that they're not compatible.
 */

#define BL_API_MIN_VERSION  1

// Every AVR is assigned a 3 byte signature sequence. The first byte is a 
// vendor ID and is the value 0x1E for Atmel. We use the remaining two bytes.

#define BL_PART_ATMEGA_16   0x9403  /**< ATMega 16                          */
#define BL_PART_ATMEGA_32   0x9502  /**< ATMega 32                          */
#define BL_PART_ATMEGA_64   0x9602  /**< ATMega 64                          */
#define BL_PART_ATMGEA_128  0x9702  /**< ATMega 128                         */

#if defined ( __AVR_ATmega16__ )
#   define  BL_PART_NUM     BL_PART_ATMEGA_16

#elif defined( __AVR_ATmega32__ )
#   define  BL_PART_NUM     BL_PART_ATMEGA_32

#elif defined( __AVR_ATmega64__ )
#   define  BL_PART_NUM     BL_PART_ATMEGA_64

#elif defined( __AVR_ATmega128__ )
#   define  BL_PART_NUM     BL_PART_ATMEGA_128

#endif

/**
 *  Maximum number of data bytes that can be trasnferred in the read/write
 *  commands. In order to remain compatible with SMBus, the maximum packet
 *  size if 32 bytes, so I picked the next lower power of 2 so that some
 *  additional data can be sent (address/size).
 */

#define BL_MAX_DATA_BYTES   16

/**
 *  Returns information in response to the BL_CMD_GET_INFO command. This
 * includes version information about the BootLoader, and some size 
 * information about the various memories.
 */

typedef struct
{
    uint8_t     version;        /**< API Version the device is using.       */
    uint8_t     minVersion;     /**< Minimum API version the device is compatible with. */
    uint16_t    partNumber;     /**< Part number (bytes 2 & 3 of signature) */
    uint16_t    regSize;        /**< Size of the register area, in bytes.   */
    uint16_t    ramSize;        /**< Size of the SRAM area, in bytes.       */
    uint16_t    eepromSize;     /**< Size of the EEPROM, in bytes.          */
    uint16_t    spmPageSize;    /**< Size of the SPM Page Size.             */
    uint32_t    flashSize;      /**< Size of the flash, in bytes.           */

} BootLoaderInfo;

/**
 *  Specifies the address and size for the BL_CMD_READ_FLASH_DATA or
 * BL_CMD_READ_EEPROM_DATA commands.
 */

typedef struct
{
    uint32_t    address;    /**< Byte address to read from.                 */
    uint8_t     numBytes;   /**< Number of bytes (up to BL_MAX_DATA_BYTES)  */
    uint8_t     pad[ 3 ];   /**< Reserved - set to zero.                    */

} BootLoaderReadCmd;

/**
 *  Returns the data read from the flash or eeprom.
 */

typedef struct
{
    uint8_t     numBytes;   /**< Number of bytes read.                      */
    uint8_t     pad[ 3 ];   /**< Reserved - set to zero.                    */
    uint8_t     data[ BL_MAX_DATA_BYTES ];  /**< Data that was read.        */

} BootLoaderReadData;

/**
 *  Specifies the data to write, for the BL_CMD_WRITE_FLASH_DATA or
 *  BL_CMD_WRITE_EEPROM_DATA command.
 *
 *  When writing to flash memory, the data must all be contained within
 *  a single page (see the spmPageSize from the BootLoaderInfo structure).
 */

typedef struct
{
    uint32_t    address;    /**< Address to write data to                   */
    uint8_t     numBytes;   /**< Number of bytes to write.                  */
    uint8_t     pad[ 3 ];   /**< Reserved. Set to zero.                     */
    uint8_t     data[ BL_MAX_DATA_BYTES ];  /**< Data to write.             */

} BootLoaderWriteCmd;

/**
 *  The write result is here primarily to act as flow control (the 
 *  Bootloader can hold the SCL line low while the memory is being
 *  programmed).
 */

typedef struct
{
    uint8_t     bytesWritten;   /**< Bytes written.                         */
    uint8_t     pad[ 3 ];       /**< For alignment.                         */

} BootLoaderWriteResult;

/**
 *  Describes the amount and layout of the EEPROM data required by the 
 *  bootloader. This will be located at the top end of EEPROM.
 * 
 *  IMPORTANT: When adding data to this structure, add it to the beginning
 *             of the structure, rather than at the end. This is because
 *             it's pinned at the end of EEPROM and by growing into lower
 *  memory, we don't necessarily have to reload the EEPROM.
 */

typedef struct
{
    uint8_t     pad;

    uint8_t     ledPort;    /**< Port to write to to toggle LED.            */
    uint8_t     ledDdr;     /**< DDR location for the LED port.             */
    uint8_t     ledMask;    /**< Bitmask to write to LED port.              */

    uint16_t    i2cSpeed;   /**< i2c Bus speed, in kHz                      */
    uint8_t     i2cAddr;    /**< i2c address (in top 7 bits format)         */
    uint8_t     structSize; /**< Size of this structure, in bytes.          */

} BootLoaderEepromData;

/**
 *  Describes the amount and layout of SRAM data required by the bootloader.
 *  This will be located at the end of SRAM.
 */

typedef struct
{
    // The packet buffer. We add 1 for the command, 1 for the messge size,
    // 1 for the PEC, and 1 pad.

    uint8_t     buf[ 36 ];  /**< Incoming/Outgoing data                     */
    uint8_t     bufIdx;     /**< Index into the buffer                      */
    uint8_t     bufSize;    /**< Size of outgoing message                   */

    uint8_t     ledPort;    /**< Port to write to toggle LED.               */
    uint8_t     ledMask;    /**< Bitmask to write to LED port.              */

} BootLoaderRamData;

/**
 *  Casues the bootloader to return information about the device.
 *
 *  Param:      None
 *  Returns:    BootLoaderInfo
 */

#define BL_CMD_GET_INFO             0xB0

/**
 *  Reads data from the flash.
 * 
 *  Param:      BootLoaderReadCmd
 *  Returns:    BootLoaderReadData
 */

#define BL_CMD_READ_FLASH_DATA      0xB1

/**
 *  Writes data into the flash.
 * 
 *  Param:      BootLoaderWriteCmd
 *  Returns:    BootLoaderWriteResult
 */

#define BL_CMD_WRITE_FLASH_DATA     0xB2

/**
 *  Reads data from the eeprom.
 * 
 *  Param:      BootLoaderReadCmd
 *  Returns:    BootLoaderReadData
 */

#define BL_CMD_READ_EEPROM_DATA     0xB3

/**
 *  Writes data into the eeprom.
 * 
 *  Param:      BootLoaderWriteCmd
 *  Returns:    BootLoaderWriteResult
 */

#define BL_CMD_WRITE_EEPROM_DATA    0xB4

/**
 *  Executes the application program which has been downloaded.
 * 
 *  Param:      None
 *  Returns:    None
 */

#define BL_CMD_RUN_APP              0xB5

#define BL_TEST_CMDS    1

#if BL_TEST_CMDS

#define BL_CMD_TEST_READ    0xC0
#define BL_CMD_TEST_WRITE   0xC1
#define BL_CMD_TEST_CALL    0xC2
#define BL_CMD_TEST_NOTIFY  0xCC

#endif

/* ---- Variable Externs ------------------------------------------------- */

/**
 *  Description of variable.
 */

/* ---- Function Prototypes ---------------------------------------------- */

/*
 *  Just include prototypes here. Put full descriptions in the .c files.
 */



/** @} */

#endif /* BOOTLOADER_H */

