###########################################################################
#
# 	mkEnv.mk
#
# 	This file contains common variable definitions used during the build.
#
###########################################################################

MK_ENV_INCLUDED = 1

#--------------------------------------------------------------------------
# 	MK_ROOT describes the path to get from the directory containing the
# 	current directory (which is usually the directory containing the makefile
# 	which included this one), to the root directory, which is typically the
# 	directory containing the rules directory.

ifeq ($(MK_ROOT),)
$(error "MK_ROOT variable hasn't been set")
endif

#--------------------------------------------------------------------------
#
# Define space to be a single space character. Using " " gets the quotes
# as well, which we don't want.
#

empty :=
space := $(empty) $(empty)

#--------------------------------------------------------------------------
#
# 	MK_OS describes the "Operating system" which is the source code is 
# 	currently being compiled for. For each supported OS, there is another
# 	rules file in the $(MK_ROOT)/rules directory which is named using
# 	mkRules_$(MK_OS).mk
#

MK_SUPPORTED_OS = avr cygwin gumstix

ifeq ($(filter $(MK_OS),$(MK_SUPPORTED_OS)),)
$(warning Unsupported value for MK_OS: '$(MK_OS)')
$(warning MK_OS must be one of the following:)
$(warning $(space))
$(foreach os,$(MK_SUPPORTED_OS),$(warning $(space)  $(os)))
$(warning $(space))
$(warning Please note that these are case sensitive)
$(warning $(space))
$(error Run 'make MK_OS=DesiredOS' to correct)
endif
  
#--------------------------------------------------------------------------
#
# 	Run make with v=1 or verbose=1 to get verbose output
#

ifeq ($(v),)
export verbose = 0
else
export verbose = 1
endif

ifeq ($(verbose),)
export verbose = 0
endif

ifeq ($(verbose),0)
	Q = @
	MAKEFLAGS += -s
else
	Q =
endif
export Q

#--------------------------------------------------------------------------
#
# 	Run make with debug=1 to have non-optimized generated.
#

ifeq ($(debug),)
export debug = 1
endif

ifeq ($(debug),1)
MK_OPTIMIZE    	= 1
MK_REL_EXT		= Dbg
else
MK_OPTIMIZE    	= 0
MK_REL_EXT		= Rel
endif

MK_ADD_SYMBOLS 	= 1

#--------------------------------------------------------------------------
#
# 	MK_OBJ_DIR describes the location where the generated object and 
# 	dependancy files will be placed. 
#

#MK_OBJ_DIR := ./obj/$(MK_OS)-$(MK_REL_EXT)
MK_OBJ_DIR := ./obj

#--------------------------------------------------------------------------
#
# 	MK_LIB_DIR describes the location where generated libraries files will
# 	be placed.
#

#MK_LIB_DIR := ./lib/$(MK_OS)-$(MK_REL_EXT)
MK_LIB_DIR := ./lib

#--------------------------------------------------------------------------
#
# 	MK_EXE_DIR describes the location where generated libraries files will
# 	be placed.
#

#MK_EXE_DIR := ./exe/$(MK_OS)-$(MK_REL_EXT)
MK_EXE_DIR := ./exe

#--------------------------------------------------------------------------
#
# 	MK_RULES_DIR describes the location where all of the support makefiles
#	live.
#

MK_RULES_DIR := $(MK_ROOT)/rules

#--------------------------------------------------------------------------
#
# 	The MK_xxx_NAME macro can be used to construct the name of a library,
#	executable, or object file, by providing the basename.
#
# 	For example:
#
#  		STR_LIB_NAME = $(call MK_LIB_NAME,str)
#
# 	will assign STR_LIB_NAME with the full version of the 'str' library.
#

MK_LIB_NAME = $(foreach libBase,$(1),$(MK_LIB_DIR)/$(MK_LIB_PREFIX)$(libBase)$(MK_LIB_EXT))
MK_EXE_NAME = $(foreach exeBase,$(1),$(MK_EXE_DIR)/$(exeBase)$(MK_EXE_EXT))
MK_HEX_NAME = $(foreach hexBase,$(1),$(MK_EXE_DIR)/$(hexBase)$(MK_HEX_EXT))
MK_OBJ_NAME = $(foreach objBase,$(1),$(MK_OBJ_DIR)/$(objBase)$(MK_OBJ_EXT))

#--------------------------------------------------------------------------
#
# Setup variables for the various utilities that are required.
# 

ECHO	= @echo
MKDIR	= mkdir
RM		= rm

#--------------------------------------------------------------------------
#
# Setup stuff for the Common directory
#

MK_COMMON_DIR = $(MK_ROOT)/Common

CPPFLAGS += -I . -I $(MK_COMMON_DIR)

vpath %.c   $(MK_COMMON_DIR)
vpath %.cpp $(MK_COMMON_DIR)

#--------------------------------------------------------------------------
#
# Get OS specific environment makefile.
#

MK_OS_ENV_FILE = $(MK_RULES_DIR)/mkEnv-$(MK_OS).mk
include $(MK_OS_ENV_FILE)

