###########################################################################
#
#	mkRules.mk
#
#	This file is the common file which includes all of the rules
#	required to build.
#
###########################################################################

ifneq ($(MK_ENV_INCLUDED),1)
$(error mkEnv.mk needs to be included BEFORE mkRules.mk)
endif

#--------------------------------------------------------------------------
#
#	You can run arbitrary commands from within the context that is setup by
# 	the makefile by using:
#
#  		make exec="command to execute"
#  
# 	Note that we're also specifying the default target here.
#
# 	You can also specify arbitrary targets by using the target=foo
# 	This is useful when trying to build preprocessor (%.pp) and such.
#

ifeq ($(exec),)
ifeq ($(target),)
all: target
else
MAKECMDGOALS += $(target)
export target
all: $(target)
endif
else
MAKECMDGOALS += exec
export exec
all: exec
endif
.PHONY: all target exec

all: $(filter %/,$(MK_SRC_FILES))

#--------------------------------------------------------------------------
#
# 	Figure out what files need to be built. Note that even though all of 
#	the components of MK_LIB_NAME haven't been defined yet, this is fine 
#	because of the deferred way that make evaluates things.
#
# 	The "EXTRA" versions are files that need to be compiled, but NOT 
#	added to a library. Typically this is used when mixing libraries and 
#	executables in the same makefile.
#
# 	We use the $(addsuffix $(MK_OBJ_EXT),$(basename ...)) so that we can
#	get .c and .cpp files.
#

MK_OBJ_FILES = $(strip $(addprefix $(MK_OBJ_DIR)/, $(addsuffix $(MK_OBJ_EXT),$(basename $(filter-out %/,$(MK_SRC_FILES))))))
MK_EXTRA_OBJ_FILES = $(strip $(addprefix $(MK_OBJ_DIR)/, $(addsuffix $(MK_OBJ_EXT),$(basename $(filter-out,%/,$(MK_EXTRA_SRC_FILES))))))

MK_DEP_FILES = $(strip $(MK_OBJ_FILES:$(MK_OBJ_EXT)=.d) $(MK_EXTRA_DEP_FILES))
MK_EXTRA_DEP_FILES = $(MK_EXTRA_OBJ_FILES:$(MK_OBJ_EXT)=.d)

#--------------------------------------------------------------------------
#
#	If the calling makefile is building a library, then it should define
#	MK_LIB_BASE, and a rule to build the library will automatically be 
#	defined.
#

ifneq ($(MK_LIB_BASE),)
MK_LIB_NAME  = $(call MK_MAKE_LIB_NAME,$(MK_LIB_BASE))
endif

#--------------------------------------------------------------------------
#
# Get the OS specific stuff
#

MK_OS_RULES_FILE = $(MK_RULES_DIR)/mkRules-$(MK_OS).mk
include $(MK_OS_RULES_FILE)

#--------------------------------------------------------------------------
#
# 	Since MK_LIB_NAME can't be evaluated until the OS specific rules have
# 	been included, we need to put the targets that use MK_LIB_NAME after
# 	the include.
#

ifneq ($(MK_LIB_NAME),)
lib_target: $(MK_LIB_NAME)
target: lib_target
.PHONY: lib_target
endif

ifneq ($(MK_EXE_TARGET),)
target: exe_target
exe_target: $(MK_EXE_TARGET)
.PHONY: exe_target
endif

ifneq ($(MK_HEX_TARGET),)
target: hex_target
hex_target: $(MK_HEX_TARGET)
.PHONY: hex_target
endif

#--------------------------------------------------------------------------
#
#	The FORCE target can be used for things that you want to have rebuild
# 	every single time (like *.pp files)
#

FORCE:

.PHONY: FORCE

#--------------------------------------------------------------------------
#
# Include dependencies.
#
# The '-' at the beginning tells make not to complain about missing
# dependency files.
#

ifneq ($(MK_DEP_FILES),)
ifeq ($(strip $(filter clean% exec print-%, $(MAKECMDGOALS))),)
-include $(MK_DEP_FILES)
endif
endif

#--------------------------------------------------------------------------
#
# 	You can run arbitrary commands from within the context that is setup 
#	by the makefile by using:
#
#  make exec="command to execute"
#  

exec:
	 @$(ECHO) "About to execute '$(exec)' ..."
	 $(exec)

#--------------------------------------------------------------------------
#
#	Targets for cleaning stuff up
#

clean: clean_dirs clean_obj clean_dep clean_lib clean_exe

clean_dirs: $(filter %/,$(MK_SRC_FILES))
				  
clean_dep:
clean_exe:
clean_lib:
clean_obj:

MK_CLEAN_DESCR = $(strip $(MK_LIB_BASE) $(MK_EXE_TARGET) $(MK_HEX_TARGET))

ifneq ($(MK_DEP_FILES),)
clean_dep: mk_clean_dep
mk_clean_dep:
	 @$(ECHO) "Removing $(MK_CLEAN_DESCR) dependencies ..."
	 $(Q)$(RM) -f $(MK_DEP_FILES)
	 
.PHONY: mk_clean_dep
endif

ifneq ($(MK_OBJ_FILES),)
clean_obj: mk_clean_obj
mk_clean_obj:
	 @$(ECHO) "Removing $(MK_CLEAN_DESCR) objects ..."
	 $(Q)$(RM) -f $(MK_OBJ_FILES)
	 
.PHONY: mk_clean_obj
endif

# Note: clean_exe and clean_lib are found in the OS specific rules files

.PHONY: clean clean_dirs clean_obj clean_dep clean_lib clean_exe

#--------------------------------------------------------------------------
#
# 	Print out a bunch of useful information
#
# 	Note that print-include is provided within the OS specific rules file.
#

.PHONY: print-info print-cfg print-include print-path

print-info: print-cfg print-include print-path

print-cfg:
	 @$(ECHO)
	 @$(ECHO) "MK_OS        = $(MK_OS)"
	 @$(ECHO) "MK_ROOT      = $(MK_ROOT)"
	 @$(ECHO) "MK_OBJ_DIR   = $(MK_OBJ_DIR)"
	 @$(ECHO) "MK_LIB_DIR   = $(MK_LIB_DIR)"
	 @$(ECHO) "MK_EXE_DIR   = $(MK_EXE_DIR)"
	 @$(ECHO) "MK_RULES_DIR = $(MK_RULES_DIR)"
	 @$(ECHO)

print-path:
	 @$(ECHO)
	 @$(ECHO) -e "PATH =\n  $(subst :,\n  ,$(PATH))"
	 @$(ECHO)

print-lib:
	 @$(ECHO) "MK_LIB_BASE = $(MK_LIB_BASE)"
	 @$(ECHO) "MK_LIB_NAME = $(MK_LIB_NAME)"
	 
print-src:
	 @$(ECHO) "MK_SRC_FILES = $(MK_SRC_FILES)"

print-obj:
	 @$(ECHO) "MK_OBJ_FILES = $(MK_OBJ_FILES)"

print-dep:
	 @$(ECHO) "MK_DEP_FILES = $(MK_DEP_FILES)"

#--------------------------------------------------------------------------
#
# 	Rule to build subdirectories
#

%/: FORCE
	$(Q)$(MAKE) -C $@ $(filter clean% exec print-%,$(MAKECMDGOALS))

